/***********************************************************************
SOFTWARE RELEASE: MarMOT 1.1.1 DATE: 23 January 2015 LICENCE: CeCILL 2.1

Copyright 2015 John C. P. Hemmings 

Includes functions 'randominit' and 'randomu' (WELL1024a versions) from
the 2013 MarMOT 1.1 release (marMOT 1.1.0)
 - Copyright 2013 Natural Environment Research Council, UK 

Author: John Hemmings
e-mail: marmot@wessexenv.mail1.co.uk 

This software is governed by the CeCILL  license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.

***********************************************************************/
/* garandom.c - John Hemmings - 09 Jan 2015 */

#include <stdlib.h>
#include <stdio.h>
#include <limits.h>
#include "garandom.h"

#ifdef no_well1024a 
  #define RANDOM_GEN 0 
#else
  #define RANDOM_GEN 1 
#endif 

/* -------------------------------------------------------------------------- */
#if RANDOM_GEN == 1 

/* Wrapper for the WELL1024a (1024 bit well-equidistributed long-period linear) 
   pseudo random number generator from
   F. Panneton, P. L'Ecuyer, and M. Matsumoto (2006)
   Improved Long-Period Generators Based on Linear Recurrences Modulo 2, 
   ACM Transactions on Mathematical Software, 32, 1-16. */ 


#include "well1024a.h"


char *randomdescrip (int *bannerflag)

/* Returns a description of the random number generator and a flag indicating
   whether it should be displayed at startup of the main program (e.g. as a 
   copyright notice) */

{ *bannerflag = 1;  

  return 
  " WELL1024a random number generator - Copyright 2005 Francois Panneton & Pierre\n L'Ecuyer, University of Montreal and Makoto Matsumoto, Hiroshima University\n";
}


void randominit (unsigned int seed)

/* Initializes the generator state */

{ unsigned int initstate[32];
  int i;

  if (UINT_MAX != 4294967295U)
    { fprintf(stderr,
"*** Error in 'randominit': WELLRNG1024a generator not compatible with system - requires 32-bit unsigned integers ***\n");
      abort();
    } 
  srand(seed); 
  for (i = 0; i < 32; i++) initstate[i] = rand();
  InitWELLRNG1024a(initstate);
}


double randomu (int printflag)

/* Returns a uniform random variate in the range 0-1 */

{ double x;

  x = WELLRNG1024a();
  if (printflag) printf("Uniform random variate: randomu = %g\n", x);
  return x;
}


/* -------------------------------------------------------------------------- */
#elif RANDOM_GEN == 0 

/* Wrapper for random number generator provided by 'stdlib' */


char *randomdescrip (int *bannerflag)

/* Returns a description of the random number generator and a flag indicating
   whether it should be displayed at startup of the main program (e.g. as a 
   copyright notice) */

{ *bannerflag = 0;

  return 
" Basic random number generation using C standard library function 'rand'\n";
}


void randominit (unsigned int seed)

/* Initializes the generator state */

{ srand(seed);
}


double randomu (int printflag)

/* Returns a uniform random variate in the range 0..1 */

{ double x;

  x = (double) rand() / RAND_MAX;
  if (printflag) printf("Uniform random variate: randomu = %g\n",x);
  return x;
}


/* -------------------------------------------------------------------------- */
#endif


int ga_random_int (int min, int max, int printflag)

/* Return random signed integer in the range min..max
   (Distribution of variates returned by 'randomu' assumed to be in
   the range 0..1, else resulting distribution is modified due to truncation) */

{ double x;

  if (min >= max) return min;

  x = ((double)min + ((double)(max - min) + 1.0) * randomu(printflag));

  if (x >= (double)max) return max;
  if (x <= (double)min) return min;

  if (x >= 0.0) return (int)x;
  else return (int)(x - 1.0);
} 


unsigned long ga_random_ulong (unsigned long min, unsigned long max, 
                               int printflag)

/* Return random unsigned long integer in the range min..max
   (Distribution of variates returned by 'randomu' assumed to be in
   the range 0..1, else resulting distribution is modified due to truncation) */

{ double x;

  if (min >= max) return min;

  x = (double)min + ((double)(max - min) + 1.0) * randomu(printflag); 

  if (x >= (double)max) return max;
  if (x <= (double)min) return min;
 
  return (unsigned long)x;
}
