/***********************************************************************
SOFTWARE RELEASE: MarMOT 1.1.1 DATE: 23 January 2015 LICENCE: CeCILL 2.1

Includes modifications to 2013 MarMOT 1.1 code (MarMOT 1.1.0) as listed 
 - Copyright 2015 John C. P. Hemmings 
***********************************************************************/
/***********************************************************************

SOFTWARE RELEASE: MarMOT 1.1  DATE: 21 November 2013

Copyright Natural Environment Research Council, UK

Author: John Hemmings, National Oceanography Centre 
e-mail: j.hemmings@noc.ac.uk (until 30 Nov 2013) 
Alternative e-mail for author: marmot@wessexenv.mail1.co.uk 

This software is governed by the CeCILL  license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.

***********************************************************************/
/***********************************************************************
 
 gfan.h - John Hemmings - 17 Feb 2013 

 MODIFIED:

 John Hemmings - 09 Jan 2015
 - Added default definition of OBJF_FREEPARMTRANSFORM_REQ 
 - Added support for code to be built without the NetCDF library 
 - Remove elitism flag from GA control record ('gactrlrec'); option for
   no elitism was unsuitable for use with the micro GA and is now
   disabled

***********************************************************************/


#include "toolbox.h"
#include "objfun.h"
#include <time.h>

#ifdef no_netcdf
  #define NETCDF_ENABLED 0
#else
  #define NETCDF_ENABLED 1
#endif

#define PAGEWIDTH 80
#define NPAGEREC_MAX 50
#define CWIDTH_MAX LINE_MAX
#define COLSEP_LOG COLSEP_LINE


/* Use default application definitions if not specified in objfun.h */

#if !defined(N_DIM)
#define N_DIM 0
#endif

#if !defined(N_CASEVAR)
#define N_CASEVAR 0
#endif

#if !defined(N_PSET)
#define N_PSET 0
#endif

#if !defined(N_MATITEM)
#define N_MATITEM 0
#endif

#if !defined(N_GDOMITEM)
#define N_GDOMITEM 0
#endif

#if !defined(N_NGDOMITEM)
#define N_NGDOMITEM 0
#endif

#if !defined(N_ITEMVAR_MAX)
#define N_ITEMVAR_MAX 1
#endif

#if !defined(N_OUTTAB)
#define N_OUTTAB 0
#endif

#if !defined(OBJF_FLOAT_TYPE) 
#define OBJF_FLOAT_TYPE float
#endif

#if !defined(OBJF_SIGDIG_DEF)
#define OBJF_SIGDIG_DEF FLT_DIG 
#endif

#if !defined(OBJF_NODATA)
#define OBJF_NODATA -9999.0 
#endif

#if !defined(OBJF_SETUP_REQ)
#define OBJF_SETUP_REQ 0
#endif

#if !defined(OBJF_INPUTCHECK_REQ) 
#define OBJF_INPUTCHECK_REQ 0
#endif

#if !defined(OPT_ENABLED) || N_PSET == 0
#define OPT_ENABLED 0
#endif

#define N_ITEM N_PSET + N_MATITEM + N_GDOMITEM + N_NGDOMITEM

#if N_CASEVAR == 0
#define N_CASEVAR_NZ 1
#else
#define N_CASEVAR_NZ N_CASEVAR
#endif

#if N_DIM == 0
#define N_DIM_NZ 1
#else
#define N_DIM_NZ N_DIM
#endif

#if N_OUTTAB == 0
#define N_OUTTAB_NZ 1
#else
#define N_OUTTAB_NZ N_OUTTAB
#endif

#if N_ITEM == 0
#define N_ITEM_NZ 1
#else
#define N_ITEM_NZ N_ITEM
#endif


/* Optimizer-related definitions */

#if OPT_ENABLED == 1

#if !defined(N_PARMCTRLVAL)
#define N_PARMCTRLVAL 0
#endif

#if !defined(N_FREEPARM_MAX)
#define N_FREEPARM_MAX 100
#endif

#if !defined(OBJF_FREEPARMCHECK_REQ)
#define OBJF_FREEPARMCHECK_REQ 0
#endif

#if !defined(OBJF_FREEPARMTRANSFORM_REQ)
#define OBJF_FREEPARMTRANSFORM_REQ 0
#endif

#if N_FREEPARM_MAX == 0
#define N_FREEPARM_MAX_NZ 1
#else
#define N_FREEPARM_MAX_NZ N_FREEPARM_MAX
#endif

#if N_PARMCTRLVAL == 0
#define N_PARMCTRLVAL_NZ 1
#else
#define N_PARMCTRLVAL_NZ N_PARMCTRLVAL
#endif

#endif


/*** OBJECTIVE FUNCTION NAME RECORD ***/

struct itemnamerec {
  char *name;
  unsigned char optional;
  unsigned char dimenabled[N_DIM_NZ];
  unsigned char dimoptional[N_DIM_NZ];
  int nvar;
  char *varname[N_ITEMVAR_MAX];
  char *vartext[N_ITEMVAR_MAX];
  unsigned char varoptional[N_ITEMVAR_MAX];
  OBJF_FLOAT_TYPE varmin[N_ITEMVAR_MAX];
  OBJF_FLOAT_TYPE varmax[N_ITEMVAR_MAX];
};

struct outtabnamerec {                     
  char *name;
  int nvar;
  char *varname[NFIELD_MAX];               
  char *vartext[NFIELD_MAX];
};

struct objfnamerec {
  char *function;
  char *version;
  char *description;
  char *notes;
  char *dimension[N_DIM_NZ];
  char *casevar[N_CASEVAR_NZ];
  struct itemnamerec item[N_ITEM_NZ];
  struct outtabnamerec outtab[N_OUTTAB_NZ];
#if OPT_ENABLED == 1
  char *parmctrlvar[N_PARMCTRLVAL_NZ];
#endif
};


/*** ITEM TABLE ***/

enum itemcategory {PSET,MATRIX,GDOMITEM,NGDOMITEM};

struct gridspecrec {
  int iregflag;
  OBJF_FLOAT_TYPE first;
  OBJF_FLOAT_TYPE last;
  OBJF_FLOAT_TYPE step;
  int n;
  OBJF_FLOAT_TYPE min;
  OBJF_FLOAT_TYPE max;
  OBJF_FLOAT_TYPE rsearch;
};

struct itemrec {
  char *fieldlist;
  char *caseid[N_CASEVAR_NZ];
  char *itemkey;
  int nrow;                              /* Matrix items only */
  int ncol;                              /* Matrix items only */
  int dimflag[N_DIM_NZ];                 /* Domain items only */
  struct gridspecrec gspec[N_DIM_NZ];    /* Gridded domain items only */
  int npoint;                            /* Non-gridded domain items only */
  OBJF_FLOAT_TYPE *parm;                 /* Parameter set items only */ 
  OBJF_FLOAT_TYPE *dim[N_DIM_NZ];        /* Active dimensions in non-gridded 
                                              domain items only */
  OBJF_FLOAT_TYPE *value[N_ITEMVAR_MAX]; /* Active variables only */ 
  int *count[N_ITEMVAR_MAX];             /* Items from NetCDF data sets only */   int objfunflag;
};

struct idspecrec {
  int itemkey;
  int ncvar;
  int cvar[N_CASEVAR_NZ];
  char *varname[N_CASEVAR+1];
};

enum gspecmodecode {STEP_AND_N,LAST_AND_N,LAST_AND_STEP,
                    IREG_STEP_AND_N,IREG_MAX_AND_N,IREG_MAX_AND_STEP};

struct itemtabrec {
  struct idspecrec idspec;
  enum itemcategory itemcat; 
  int dimflag[N_DIM_NZ];                  /* Domain items only */
  enum gspecmodecode gspecmode[N_DIM_NZ]; /* Gridded domain items only */
  unsigned char varactive[N_ITEMVAR_MAX];
  int ndata;
  struct itemrec *table;
};


/*** CASE TABLE ***/

struct caseinputrec {
  int casenum;
  char *fieldlist;
  char *caseid[N_CASEVAR_NZ];
  char *itemkey[N_ITEM_NZ];
  int itemnum[N_ITEM_NZ];
  struct itemrec *item[N_ITEM_NZ];
};

struct casetabrec {
  struct objfnamerec *objfname; 
  void *fixeddata;
  int optpsetnum;
  int keyflag[N_ITEM_NZ];
  unsigned char itemactive[N_ITEM_NZ];
  int ndata; 
  struct caseinputrec *table;
};


#if OPT_ENABLED == 1

struct gactrlrec {
  int maxgen;            /* Max. no. of generations */
  int mingen;            /* Min. no. of generations */
  int npopsiz;           /* Parameter vector population size */
  int nout;              /* No. of output parameter vectors */
  int nchild;            /* Children per parent pair */
  int psetini;           /* Explicit initialization flag */
  int iunifrm;           /* Uniform crossover flag */
  double pcross;         /* Crossover probability */
  int seed;              /* Random number seed */
};

struct powellctrlrec {
  double ffrctol;        /* Fractional tolerance for fn. value */
  double xabstol;        /* Absolute parameter vector tolerance
                            in space seen by Powell algorithm */
  double ffrctolbrent;   /* Fractional tol. for fn. in Brent */
  double xfrctolbrent;   /* Tol. parameter for abcissa in Brent */
  double xabstolbrent;   /* Tol. parameter for abcissa in Brent */
  int maxiter;           /* Max. Powell iterations */
  int maxiterbracket;    /* Max. bracketing algorithm iterations */
  int maxiterbracket0;   /* Max. bracketing algorithm iterations
                            for finding a slope */
  int maxiterbrent;      /* Max. Brent iterations */
};


/* Free parameter control arrays */

struct parmctrlrec {
  int nparm;
  int parmno[N_FREEPARM_MAX_NZ];
  int parminitindex[N_FREEPARM_MAX_NZ];
  int logtrans[N_FREEPARM_MAX_NZ];
  int finitepowell[N_FREEPARM_MAX_NZ];
  OBJF_FLOAT_TYPE logmin[N_FREEPARM_MAX_NZ];
  OBJF_FLOAT_TYPE logmax[N_FREEPARM_MAX_NZ];
  OBJF_FLOAT_TYPE min[N_FREEPARM_MAX_NZ];
  OBJF_FLOAT_TYPE max[N_FREEPARM_MAX_NZ];
  int nbits[N_FREEPARM_MAX_NZ];
  OBJF_FLOAT_TYPE step[N_FREEPARM_MAX_NZ];
  OBJF_FLOAT_TYPE value[N_PARMCTRLVAL_NZ][N_FREEPARM_MAX_NZ];
};


/* Free parameter initial value table */

struct psetrec {
  int psetno;
  OBJF_FLOAT_TYPE *parm;
};

struct psettabrec {
  int ndata;
  int nparm;
  int parmno[N_FREEPARM_MAX_NZ];
  struct psetrec *table;
};


/* Report data for new parameter set */

struct optparmrec {
  double funcmin;
  int psetno;
  OBJF_FLOAT_TYPE parm[N_FREEPARM_MAX_NZ];
  int itercount;
  long funccount;
  int status;
};


/* Optimizer data area */

struct optdatarec {
  struct gactrlrec gactrl;
  struct powellctrlrec powellctrl;
  struct parmctrlrec freeparm;
  struct psettabrec parminittab;
  struct psettabrec parminittab2;
  struct optparmrec report;
  struct casetabrec *funcindata;
  FILE *funcoutfile[N_OUTTAB_NZ];
};

#endif


/*** FUNCTION PROTOTYPES ***/


#if OPT_ENABLED == 1

/* GFAn optimizer function */

int optimizeparameters (FILE *, int, char *, FILE *, FILE *, FILE *, 
                        FILE *, int, struct optdatarec *, struct objfnamerec *, 
                        char **);

/* GFAn utility functions used by optimizer */

int writeobjfevaldata_ga (FILE *, int, int, int, int, OBJF_FLOAT_TYPE *,
                          double, double);
int writeobjfevaldata_powell (FILE *, int, int, int, int, int,
                              int, OBJF_FLOAT_TYPE *, double, double);
int writenewparmdata_noopt (FILE *, struct parmctrlrec *,
                            struct psettabrec *, double *);
int writenewparmdatarec (FILE *, struct parmctrlrec *,
                         struct optparmrec *, clock_t, double);
int writelog (FILE *, int, char *);
int writeheading (FILE *, char *);
int writelogfreeparmtab (FILE *, struct parmctrlrec *,
                         OBJF_FLOAT_TYPE [], char *[], char *[]);
int writelogparminittab (FILE *, struct psettabrec *, char *[]);

#endif


/* Standard "Objective function" functions called by GFAn */

double objfun (struct casetabrec *, FILE *[], int *, char **);
int objfun_setnames (struct objfnamerec *, char **);

#if OBJF_SETUP_REQ == 1 
int objfun_setup (struct casetabrec *);
#endif

#if OBJF_INPUTCHECK_REQ == 1 
int objfun_inputcheck (struct casetabrec *, char **);
#endif

/* GFAn utility functions for use by objective function */

int writeouttabvardescrip (int, FILE *outfile, struct outtabnamerec *, char *,
                           int);
int writeouttabvar (int, FILE *, char *[], struct outtabnamerec *);
int writeouttabdata (int, FILE *, int, int, struct caseinputrec *, int, ...);
int writeouttabdata_array (int, FILE *, int, int, struct caseinputrec *, int, 
                           OBJF_FLOAT_TYPE *);


/*** WORK AREA FOR PAGE OUTPUT ***/

struct datarec dpage[NPAGEREC_MAX];

