/***********************************************************************
SOFTWARE RELEASE: MarMOT 1.1.1 DATE: 23 January 2015 LICENCE: CeCILL 2.1

Includes modifications to 2013 MarMOT 1.1 code (MarMOT 1.1.0) as listed 
***********************************************************************/
/***********************************************************************

SOFTWARE RELEASE: MarMOT 1.1  DATE: 21 November 2013

Copyright Natural Environment Research Council, UK

Author: John Hemmings, National Oceanography Centre 
e-mail: j.hemmings@noc.ac.uk (until 30 Nov 2013) 
Alternative e-mail for author: marmot@wessexenv.mail1.co.uk 

This software is governed by the CeCILL  license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.

***********************************************************************/
/***********************************************************************

 optimizer.c - John Hemmings - 29 Jan 2013

 MODIFIED:

 John Hemmings - 24 Nov 2014
 - Included prototypes for 'objfun_setknobs' and 'objfun_setparmvals'
 functions previously declared in 'optimizer.h'
 - Removed memory allocation and release for G.A. internal workspace 
 (now handled by the 'minimize_ga' function)
 
***********************************************************************/

#include "optimizer.h"

static void quicksortpsets (int [], double [], int, int);
static void swap (int [], int, int);
static int optimizelocal (int, struct optdatarec *, FILE *, FILE *, FILE *, 
                          char *[], char **);
static double objfdriver_ga (struct gadatarec *, OBJF_FLOAT_TYPE []);
static double objfdriver_powell (struct powelldatarec *, double []);
static void updobjfparm (int, struct parmctrlrec *, OBJF_FLOAT_TYPE [],
                         struct casetabrec *);

#if OBJF_FREEPARMTRANSFORM_REQ == 0

int objfun_setknobs (struct parmctrlrec *, OBJF_FLOAT_TYPE [], double [],
                     char**);
int objfun_setparmvals (struct parmctrlrec *, double [], OBJF_FLOAT_TYPE [],
                        char **);

#endif


int optimizeparameters (FILE *logfile, int echo, char *linebuff, 
                        FILE *optreportfile, FILE *funcgafile,
                        FILE *funcpowellfile, FILE *newparmfile,
                        int ioptpset, struct optdatarec *optdata, 
                        struct objfnamerec *objfname, char **errtxt)

{ struct gadatarec gadata;        /* Data area for G.A. */
  OBJF_FLOAT_TYPE *parmval;       /* Population of parameter vectors */
  double *funcval;                /* Function value for each parameter vector */
  int *psetbest;                  /* Id numbers ('psetkey') of param. vectors */
  struct psettabrec *pinittab2;   /* Table of parameter vectors for 
                                     local optimization */
  int nparm,i,j,k,status;
  OBJF_FLOAT_TYPE *parmitem;
  char *parmname[N_FREEPARM_MAX_NZ];

  *errtxt=NULL;
  nparm=optdata->freeparm.nparm;

  for (i=0;i<nparm;i++)
    { j=optdata->freeparm.parmno[i]-1;
      parmname[i]=objfname->item[ioptpset].varname[j];
    }

  pinittab2=&optdata->parminittab2;
  pinittab2->table=NULL;
  parmval=NULL;
  funcval=NULL;
  psetbest=NULL;

  if (optdata->gactrl.maxgen>0)

    { /* Search for minima globally using Carroll's genetic algorithm */

      /* Allocate G.A. interface workspace */

      parmval=(OBJF_FLOAT_TYPE *)
        malloc(optdata->gactrl.npopsiz*nparm*sizeof(OBJF_FLOAT_TYPE));
      if (!parmval) return MEMERR;

      funcval=(double *)malloc(optdata->gactrl.npopsiz*sizeof(double));
      if (!funcval) return MEMERR;

      psetbest=(int *)malloc(optdata->gactrl.npopsiz*sizeof(int));
      if (!psetbest) return MEMERR;

      /* Initialize G.A. parameter vector population (if applicable) */ 

      if (optdata->gactrl.psetini==1)

        { for (i=0;i<optdata->gactrl.npopsiz;i++)
            { for (j=0;j<nparm;j++)
                { k=optdata->freeparm.parminitindex[j];
                  parmval[i*nparm+j]=optdata->parminittab.table[i].parm[k];
                }
            }

          status=writelog(logfile,echo,
                          "Starting G.A. optimization (external init.)");
        }

      else
        status=writelog(logfile,echo,"Starting G.A. optimization");

      if (status!=OK) return status;

      /* Run the G.A. 
          - returned function values are fitness values not cost values, 
            so are negated */

      gadata.parmname=parmname;
      gadata.func=&objfdriver_ga;
      gadata.ctrl=&optdata->gactrl;
      gadata.logfile=optreportfile;
      gadata.parmctrl=&optdata->freeparm;
      gadata.funcdata=optdata->funcindata;
      gadata.ipset=ioptpset;
      gadata.evalfile=funcgafile;
      for (i=0;i<N_OUTTAB;i++) gadata.funcoutfile[i]=NULL;
      gadata.startcpu=clock();

      status=minimize_ga(&gadata,parmval,funcval,psetbest); 
      if (status!=OK) return status;

      /* Set up parameter initial value table for use by Powell optimizer
         (or for direct output to new parameter table in the absence of 
         local optimization).
         The table comprises the requested number of parameter sets
         from the top of a merit list.
         The merit list is created by sorting the G.A. output population
         of parameter sets by the objective function value.
         The parameter set keys in the table refer to the parameter
         sets in the unsorted population. */
     
      for (i=0;i<optdata->gactrl.npopsiz;i++)
        { funcval[i]=-funcval[i];
                      /* Convert fitness to objective function value */
          psetbest[i]=i;
        }
      quicksortpsets(psetbest,funcval,0,optdata->gactrl.npopsiz-1);

      pinittab2->nparm=nparm;
      for (i=0;i<nparm;i++) pinittab2->parmno[i]=
        optdata->freeparm.parmno[i];
      pinittab2->ndata=optdata->gactrl.nout;
      pinittab2->table=
        (struct psetrec *)malloc(pinittab2->ndata*sizeof(struct psetrec));
      if (!pinittab2->table) return MEMERR;

      for (i=0;i<optdata->gactrl.nout;i++)
        { j=psetbest[i];
          pinittab2->table[i].psetno=j+1;
          pinittab2->table[i].parm=&parmval[j*nparm];
        }
    }

  if (optdata->powellctrl.maxiter>0)

    { /* Optimize locally using Powell's direction set method
         On exit from 'optimizelocal' report data relates to best
         parameter set */

      if (optdata->gactrl.maxgen>0)

        { /* Initialize local optimization from G.A. output table */

          sprintf(linebuff,
    "Starting local optimization (init. from G.A. - function minimum = %.*g)",
                  OBJF_SIGDIG_DEF,funcval[psetbest[0]]);
          writelog(logfile,echo,linebuff);
        }

      else if (optdata->parminittab.ndata>0)

        { /* Initialize local optimization from external table */ 

          writelog(logfile,echo,"Starting local optimization (external init.)");

          pinittab2->nparm=nparm;
          pinittab2->ndata=optdata->parminittab.ndata;
          for (i=0;i<nparm;i++)
            pinittab2->parmno[i]=optdata->freeparm.parmno[i];
          pinittab2->table=
            (struct psetrec *)malloc(pinittab2->ndata*sizeof(struct psetrec));
          if (!pinittab2->table) return MEMERR;
          parmval=(OBJF_FLOAT_TYPE *)
            malloc(pinittab2->ndata*nparm*sizeof(OBJF_FLOAT_TYPE));
          if (!parmval) return MEMERR;

          for (i=0;i<pinittab2->ndata;i++)
            { pinittab2->table[i].psetno=
                optdata->parminittab.table[i].psetno;
              for (j=0;j<nparm;j++)
                { k=optdata->freeparm.parminitindex[j];
                  parmval[i*nparm+j]=optdata->parminittab.table[i].parm[k];
                }
              pinittab2->table[i].parm=&parmval[i*nparm];
            }
        }

      else

        { /* Initialize local optimization from mid points of ranges specified  
             in free parameter table or from parameter item associated with
             the first (or only) case in the case table */

          writelog(logfile,echo,
                   "Starting local optimization (init. from item or range)");

          pinittab2->nparm=nparm;
          pinittab2->ndata=1;
          for (i=0;i<nparm;i++) 
            pinittab2->parmno[i]=optdata->freeparm.parmno[i];
          pinittab2->table=(struct psetrec *)malloc(sizeof(struct psetrec));
          if (!pinittab2->table) return MEMERR;
          parmval=(OBJF_FLOAT_TYPE *) malloc(nparm*sizeof(OBJF_FLOAT_TYPE));
          if (!parmval) return MEMERR;

          pinittab2->table[0].psetno=0;
          for (i=0;i<nparm;i++)
            { if (optdata->freeparm.finitepowell[i])
                { if (optdata->freeparm.logtrans[i])
                    parmval[i]=pow(10,0.5*(optdata->freeparm.logmin[i]+
                               optdata->freeparm.logmax[i]));
                  else
                    parmval[i]=0.5*(optdata->freeparm.min[i]+
                               optdata->freeparm.max[i]);
                }
              else 
                { parmitem=optdata->funcindata->table[0].item[ioptpset]->parm;
                  parmval[i]=parmitem[optdata->freeparm.parmno[i]-1];
                  if (parmval[i]==OBJF_NODATA) parmval[i]=0.0;
                }
            }
          pinittab2->table[0].parm=parmval;
        }
       
      if (logfile)
        { status=writeheading(logfile,
                   "INITIAL PARAMETER VECTOR(S) FOR LOCAL OPTIMIZATION");
          if (status!=OK) return status;
          status=writelogparminittab(logfile,pinittab2,
                                     objfname->item[ioptpset].varname);
          if (status!=OK) return status; 
          writeheading(logfile,NULL);
          fflush(logfile);
        }

      status=optimizelocal(ioptpset,optdata,newparmfile,funcpowellfile,
                           optreportfile,parmname,errtxt);
      if (status!=OK)
        { writelog(logfile,echo,*errtxt);
          return status;
        }
    }

  else if (optdata->gactrl.maxgen>0)

    { /* No local optimization so write new parameter table 
         direct from parameter initial value table created by G.A. */

      if (newparmfile)
        { status=writenewparmdata_noopt(newparmfile,&optdata->freeparm,
                                        pinittab2,funcval);
          if (status!=OK) return status;
        }

      optdata->report.funcmin=funcval[0];
      optdata->report.psetno=pinittab2->table[0].psetno;
      for (i=0;i<nparm;i++)
        optdata->report.parm[i]=pinittab2->table[0].parm[i];
      optdata->report.itercount=0;
      optdata->report.funccount=0;
    }

  /* Clear up and report optimization result to log file */

  free(parmval);
  free(funcval);
  free(psetbest);
  free(pinittab2->table); 

  status=writelog(logfile,echo,"Parameter optimization completed");
  if (status!=OK) return status; 

  if (logfile)
    { sprintf(linebuff,
        "FREE PARAMETER TABLE - OPTIMAL SET (psetkey = %d)",
              optdata->report.psetno);
      status=writeheading(logfile,linebuff);
      if (status!=OK) return status; 
      status=writelogfreeparmtab(logfile,&optdata->freeparm,
        optdata->report.parm,objfname->parmctrlvar,
        objfname->item[ioptpset].varname);
      if (status!=OK) return status; 
      writeheading(logfile,NULL);
      fflush(logfile);
    }

  /* Update parameters in the application's parameter item */

  updobjfparm(ioptpset,&optdata->freeparm,optdata->report.parm,
              optdata->funcindata);

  return OK;
}


static void quicksortpsets (int index[], double value[], int left,
                            int right)

{ int i,last;

  if (left>=right) return;
  swap(index,left,(left+right)/2);
  last=left;
  for (i=left+1;i<=right;i++)
    if (value[index[i]]<value[index[left]] ||
        (value[index[i]]==value[index[left]] && index[i]<index[left]))
      swap(index,++last,i);
  swap(index,left,last);
  quicksortpsets(index,value,left,last-1);
  quicksortpsets(index,value,last+1,right);
}


static void swap (int index[], int i, int j)

{ int temp;

  temp=index[i];
  index[i]=index[j];
  index[j]=temp;
}


static int optimizelocal
  (int ipset, struct optdatarec *optdata, FILE *outfile, FILE *evalfile, 
   FILE *optreportfile, char *parmname[], char **errtxt)

{ struct powelldatarec powelldata;
  struct optparmrec report;
  struct psettabrec *pinittab;
  struct psetrec *tabentry;
  double knob[N_FREEPARM_MAX_NZ];
  int nparm,i,j,k,status,duplicate;
  time_t t1,t2;
  clock_t cputime;

  *errtxt=NULL;
  nparm=optdata->freeparm.nparm;
  pinittab=&optdata->parminittab2;

  if (pinittab->ndata<=0)
    { *errtxt="no initial parameter sets";
      return SPECIAL;
    }

  /* Initialize new parameter report data */

  optdata->report.funcmin=FLT_MAX;
  optdata->report.psetno=0;
  for (i=0;i<nparm;i++) optdata->report.parm[i]=0;
  optdata->report.itercount=0;
  optdata->report.funccount=0;
  optdata->report.status=0;

  /* Configure Powell optimizer */

  powelldata.func=&objfdriver_powell;
  powelldata.ctrl=&optdata->powellctrl;
  powelldata.logfile=optreportfile;
  powelldata.parmctrl=&optdata->freeparm;
  powelldata.n=nparm;
  powelldata.funcdata=optdata->funcindata;
  powelldata.ipset=ipset;
  powelldata.evalfile=evalfile;
  for (i=0;i<N_OUTTAB;i++) powelldata.funcoutfile[i]=NULL;

  /* Allocate workspace */

  powelldata.direction=(double **)malloc((size_t)(nparm*sizeof(double*)));
  if (!powelldata.direction) return MEMERR;
  powelldata.direction[0]=
    (double *)malloc((size_t)(nparm*nparm*sizeof(double)));
  if (!powelldata.direction[0]) return MEMERR;
  for(i=1;i<nparm;i++) powelldata.direction[i]=powelldata.direction[i-1]+nparm;

  powelldata.dir=(double *)malloc((size_t)nparm*sizeof(double));
  if (!powelldata.dir) return MEMERR;

  powelldata.p0=(double *)malloc((size_t)nparm*sizeof(double));
  if (!powelldata.p0) return MEMERR;

  powelldata.pext=(double *)malloc((size_t)nparm*sizeof(double));
  if (!powelldata.pext) return MEMERR;

  powelldata.parmval=(double *)malloc((size_t)nparm*sizeof(double));
  if (!powelldata.parmval) return MEMERR;

  /* Start sequence of optimizations */

  duplicate=0;
  for (i=0;i<pinittab->ndata;i++)

    { /* Extract initial parameter values from table entry by parameter
         number */

      tabentry=&pinittab->table[i];
      report.psetno=tabentry->psetno;
      if (i>0)
        { duplicate=1;
          for (j=0;j<nparm;j++)
            if (tabentry->parm[j]!=pinittab->table[i-1].parm[j]) duplicate=0;
        }

      /* Optimize if parameter vector is not a duplicate */

      t1=time(NULL);
      powelldata.startcpu=clock();

      if (duplicate)

        { report.funccount=0;
          report.itercount=0;
          t2=t1;
          cputime=0;
        }

      else

        { if (optreportfile) 
            { fprintf(optreportfile,
                "\n\nINITIAL VECTOR FOR LOCAL OPTIMIZATION (psetkey=%d):\n",
                      report.psetno);
              for (j=0;j<nparm;j++) 
                fprintf(optreportfile," %s=%g",parmname[j],tabentry->parm[j]);
              fprintf(optreportfile,"\n");
              fflush(optreportfile);
             } 

          /* Convert objective function parameter values to knobs */
      
          status=objfun_setknobs(&optdata->freeparm,tabentry->parm,knob,errtxt);
          if (status!=OK) return status;

          /* Do Powell optimization */

          powelldata.psetno=report.psetno;
          powelldata.funccount=0;

          report.status=
            minimize_powell(&powelldata,knob,&report.funcmin,errtxt);

          t2=time(NULL);
          cputime=clock()-powelldata.startcpu;

          if (report.status!=0 && optreportfile) 
            { fprintf(optreportfile,
                      "\nLocal optimization failure (psetkey=%d): %s\n",
                      report.psetno,*errtxt);
              fflush(optreportfile);
            }

          report.itercount=powelldata.iternum;
          report.funccount=powelldata.funccount;
  
          /* Convert knobs to objective function parameter values */
      
          status=objfun_setparmvals(&optdata->freeparm,knob,report.parm,errtxt);
          if (status!=OK) return status;
        }

      /* Write new parameter set to file */

      if (outfile) 
        { status=writenewparmdatarec(outfile,&optdata->freeparm,&report,
                                     cputime,difftime(t2,t1));
          if (status!=OK) return status;
        }

      /* If this is the lowest function value found then save current report */

      if (report.funcmin<optdata->report.funcmin) optdata->report=report;
    }

  /* Free workspace */

  free(powelldata.direction[0]);
  free(powelldata.direction);
  free(powelldata.dir);
  free(powelldata.p0);
  free(powelldata.pext);
  free(powelldata.parmval);

  return OK;
}


static double objfdriver_ga (struct gadatarec *d, OBJF_FLOAT_TYPE parmvec[])

{ double x;
  int status;
  char *errtxt;
  clock_t t; 

  /* Copy parameter values into objective function's input data */

  updobjfparm(d->ipset,d->parmctrl,parmvec,d->funcdata);

  /* Evaluate objective function (or use a known value if available) */

  if (d->knownfitness) x=-(*d->knownfitness); 
  else
 
    { x=objfun(d->funcdata,d->funcoutfile,&status,&errtxt);

      if (status!=OK)
        { if (d->logfile)
            { fprintf(d->logfile,
                      "\nError in objective function. Optimization aborted\n");
              if (errtxt) fprintf(d->logfile,"%s\n",errtxt);
            }
          fprintf(stderr,"Error in objective function. Optimization aborted\n");
          abort();
        }
    }

  /* Write function evaluation to file */

  if (d->evalfile)
    { t=clock();
      status=writeobjfevaldata_ga(d->evalfile,1,d->gennum,d->psetnum,
                                  d->parmctrl->nparm,parmvec,x,
                                  t-d->startcpu);
      if (status!=OK)
        { if (d->logfile)
            fprintf(d->logfile,
              "\nError in writeobjfevaldata_ga. Optimization aborted\n");
          fprintf(stderr,
            "Error in writeobjfevaldata_ga. Optimization aborted\n");
          abort();
        }
    }

  return -x;
}


static double objfdriver_powell (struct powelldatarec *d, double knob[])

{ OBJF_FLOAT_TYPE parm[N_FREEPARM_MAX_NZ];
  double x;
  int status;
  char *errtxt;
  clock_t t;

  /* Convert knobs to objective function parameter values */

  status=objfun_setparmvals(d->parmctrl,knob,parm,&errtxt);

  if (status!=OK)
    { if (d->logfile)
        { fprintf(d->logfile,
                  "\nError in objfun_setparmvals. Optimization aborted\n");
          if (errtxt) fprintf(d->logfile,"%s\n",errtxt);
        }
      fprintf(stderr,"Error in objfun_setparmvals. Optimization aborted\n");
      abort();
    }

  /* Copy parameter values into objective function input data,
     overwriting default values for all cases */

  updobjfparm(d->ipset,d->parmctrl,parm,d->funcdata);

  /* Execute objective function */

  x=objfun(d->funcdata,d->funcoutfile,&status,&errtxt);

  if (status!=OK)
    { if (d->logfile)
        { fprintf(d->logfile,
                  "\nError in objective function. Optimization aborted\n");
          if (errtxt) fprintf(d->logfile,"%s\n",errtxt);
        }
      fprintf(stderr,"Error in objective function. Optimization aborted\n");
      abort();
    }

  d->funccount++;

  /* Write function evaluation to file */

  t=clock();
  d->evalnum++;

  if (d->evalfile)
    { status=writeobjfevaldata_powell(d->evalfile,1,d->psetno,d->iternum,
                                      d->dirnum,d->evalnum,d->n,parm,x,
                                      t-d->startcpu);
      if (status!=OK)
        { if (d->logfile)
            fprintf(d->logfile,
              "\nError in writeobjfevaldata_powell. Optimization aborted\n");
          fprintf(stderr,
            "Error in writeobjfevaldata_powell. Optimization aborted\n");
          abort();
        }
    }

  return x;
}


static void updobjfparm (int ipset, struct parmctrlrec *f,
                         OBJF_FLOAT_TYPE parm[], struct casetabrec *objfinput)

{ OBJF_FLOAT_TYPE *objfparm;
  int i,j;

  for (i=0;i<objfinput->ndata;i++)
    { objfparm=objfinput->table[i].item[ipset]->parm;
      for (j=0;j<f->nparm;j++) objfparm[f->parmno[j]-1]=parm[j];
    }
}


#if OBJF_FREEPARMTRANSFORM_REQ == 0

int objfun_setknobs (struct parmctrlrec *ctrl, OBJF_FLOAT_TYPE parm[],
                     double knob[], char **errtxt)

{ int i;
  double p,midpoint;

  for (i=0;i<ctrl->nparm;i++)
    { if (ctrl->logtrans[i])
        { if (ctrl->finitepowell[i])
            { midpoint=0.5*(ctrl->logmin[i]+ctrl->logmax[i]);
              p=log10(parm[i]);
              if (p>=midpoint)
                knob[i]=(p-midpoint)/(ctrl->logmax[i]-p);
              else
                knob[i]=(p-midpoint)/(p-ctrl->logmin[i]);
            }
          else knob[i]=log10(parm[i]);
        }
      else
        { if (ctrl->finitepowell[i])
            { midpoint=0.5*(ctrl->min[i]+ctrl->max[i]);
              p=parm[i];
              if (p>=midpoint)
                knob[i]=(p-midpoint)/(ctrl->max[i]-p);
              else
                knob[i]=(p-midpoint)/(p-ctrl->min[i]);
            }
          else knob[i]=parm[i];
        }
    }

  return OK;
}


int objfun_setparmvals (struct parmctrlrec *ctrl, double knob[],
                        OBJF_FLOAT_TYPE parm[], char **errtxt)

{ int i;
  double midpoint,limit;

  for (i=0;i<ctrl->nparm;i++)
    { if (ctrl->logtrans[i])
        { if (ctrl->finitepowell[i])
            { midpoint=0.5*(ctrl->logmin[i]+ctrl->logmax[i]);
              if (knob[i]>=0) limit=ctrl->logmax[i];
              else limit=ctrl->logmin[i];
              parm[i]=pow(10,(midpoint+fabs(knob[i])*limit)/(1+fabs(knob[i])));
            }
          else parm[i]=pow(10,knob[i]);
        }
      else
        { if (ctrl->finitepowell[i])
            { midpoint=0.5*(ctrl->min[i]+ctrl->max[i]);
              if (knob[i]>=0) limit=ctrl->max[i];
              else limit=ctrl->min[i];
              parm[i]=(midpoint+fabs(knob[i])*limit)/(1+fabs(knob[i]));
            }
          else parm[i]=knob[i];
        }
    }

  return OK;
}

#endif

