!**************************************************************************!
! SOFTWARE RELEASE: MarMOT 1.1.1 DATE: 23 January 2015 LICENCE: CeCILL 2.1 !
!**************************************************************************!
!**************************************************************************!
!                                                                          ! 
! SOFTWARE RELEASE: MarMOT 1.1  DATE: 21 November 2013                     ! 
!                                                                          !
! Copyright Natural Environment Research Council, UK                       !
!                                                                          !
! Author: John Hemmings, National Oceanography Centre                      !
! e-mail: j.hemmings@noc.ac.uk (until 30 Nov 2013)                         !
! Alternative e-mail for author: marmot@wessexenv.mail1.co.uk              !
!                                                                          !
! This software is governed by the CeCILL  license under French law and    !
! abiding by the rules of distribution of free software.  You can  use,    ! 
! modify and/ or redistribute the software under the terms of the CeCILL   !
! license as circulated by CEA, CNRS and INRIA at the following URL        !
! "http://www.cecill.info".                                                ! 
!                                                                          !
! As a counterpart to the access to the source code and  rights to copy,   !
! modify and redistribute granted by the license, users are provided only  !
! with a limited warranty  and the software's author,  the holder of the   !
! economic rights,  and the successive licensors  have only  limited       !
! liability.                                                               ! 
!                                                                          !
! In this respect, the user's attention is drawn to the risks associated   !
! with loading,  using,  modifying and/or developing or reproducing the    !
! software by the user in light of its specific status of free software,   !
! that may mean  that it is complicated to manipulate,  and  that  also    !
! therefore means  that it is reserved for developers  and  experienced    !
! professionals having in-depth computer knowledge. Users are therefore    !
! encouraged to load and test the software's suitability as regards their  !
! requirements in conditions enabling the security of their systems and/or ! 
! data to be ensured and,  more generally, to use and operate it in the    ! 
! same conditions as regards security.                                     ! 
!                                                                          ! 
! The fact that you are presently reading this means that you have had     !
! knowledge of the CeCILL license and that you accept its terms.           !
!                                                                          !
!**************************************************************************!
!===============================================================================
! MODULE medusa_globdata
! John Hemmings - 01 Apr 2013 
!
! Contains global MEDUSA variables
!===============================================================================
  module medusa_globdata

  use modconfig, only: nk,kind_r

  integer,parameter :: wp=kind_r

! Model configuration

! Note re arrays:
! Elements at index 1 are sometimes updated by the model but are redundant. 
! Array space is allocated to support this to avoid unnecessary changes to the
! model code.

  integer,parameter :: jpi=2    ! Array dimensions 
  integer,parameter :: jpj=2    ! 
  integer,parameter :: jpim1=2  ! Maximum index in i dimension 
  integer,parameter :: jpjm1=2  ! Maximum index in j dimension

  real(wp) fsdepw(jpi,jpj,nk)   ! Depth of top of layer (m)
  real(wp) fse3t(jpi,jpj,nk)    ! Layer thickness (m)

  integer jpk                   ! Number of model levels  

! Tracer indices

! MarMOT primary tracers
  integer,parameter :: jpphn=1  ! Non-diatoms (mmolN/m^3)
  integer,parameter :: jpphd=2  ! Diatoms (mmolN/m^3)
  integer,parameter :: jpzmi=3  ! Microzooplankton (mmolN/m^3)
  integer,parameter :: jpzme=4  ! Mesozooplankton (mmolN/m^3)
  integer,parameter :: jpdin=5  ! Dissolved inorganic nitrogen (mmolN/m^3)
  integer,parameter :: jpsil=6  ! Dissolved silicic acid (mmolSi/m^3)
  integer,parameter :: jpfer=7  ! Dissolved iron (mmolFe/m^3)
  integer,parameter :: jpdet=8  ! Detritus (mmolN/m^3)

! MarMOT derived tracers
  integer,parameter :: jpchn=9  ! Non-diatom chlorophyll (mg/m^3)
  integer,parameter :: jpchd=10 ! Diatom chlorophyll (mg/m^3)
  integer,parameter :: jppds=11 ! Diatom silicon (mmolSi/m^3)

! Model state

  integer,parameter :: nt=11    ! Number of tracers
  real(wp) trn(jpi,jpj,nk,nt)   ! Tracer concentrations
  real(wp) tra(jpi,jpj,nk,nt)   ! Tendencies

! Log output configuration

  logical lwp                   ! Switch for writing log output 
  integer,parameter :: numout=6 ! Stream for log output
  integer,parameter :: nit000=1 ! Timestep no. for log heading

! Model parameters 

!! Primary production and chl related quantities
   INTEGER  ::  jphot     !: internal photosynthesis calc. switch
   REAL(wp) ::  xxi       !: conversion factor from gC to mmolN
   REAL(wp) ::  xaln      !: Chl-a specific initial slope of P-I curve for 
                          !:   non-diatoms
   REAL(wp) ::  xald      !: Chl-a specific initial slope of P-I curve for 
                          !:   diatoms
   INTEGER  ::  jphy      !: phytoplankton T-dependent growth switch
   REAL(wp) ::  xvpn      !: maximum growth rate for non-diatoms
   REAL(wp) ::  xvpd      !: maximum growth rate for diatoms
   REAL(wp) ::  xthetam   !: maximum Chl to C ratio for non-diatoms
   REAL(wp) ::  xthetamd  !: maximum Chl to C ratio for diatoms

!! Diatom silicon parameters
   REAL(wp) ::  xsin0     !: minimum diatom Si:N ratio
   REAL(wp) ::  xnsi0     !: minimum diatom N:Si ratio
   REAL(wp) ::  xuif      !: hypothetical growth ratio at infinite Si:N ratio

!! Nutrient limitation
   REAL(wp) ::  xnln      !: half-sat constant for DIN uptake by non-diatoms
   REAL(wp) ::  xnld      !: half-sat constant for DIN uptake by diatoms
   REAL(wp) ::  xsld      !: half-sat constant for Si uptake by diatoms
   REAL(wp) ::  xfln      !: half-sat constant for Fe uptake by non-diatoms
   REAL(wp) ::  xfld      !: half-sat constant for Fe uptake by diatoms

!! Grazing
   REAL(wp) ::  xgmi      !: microzoo maximum growth rate
   REAL(wp) ::  xgme      !: mesozoo maximum growth rate
   REAL(wp) ::  xkmi      !: microzoo grazing half-sat parameter
   REAL(wp) ::  xkme      !: mesozoo grazing half-sat parameter
   REAL(wp) ::  xphi      !: micro/mesozoo grazing inefficiency
   REAL(wp) ::  xbetan    !: micro/mesozoo N assimilation efficiency
   REAL(wp) ::  xbetac    !: micro/mesozoo C assimilation efficiency
   REAL(wp) ::  xkc       !: micro/mesozoo net C growth efficiency
   REAL(wp) ::  xpmipn    !: grazing preference of microzoo for non-diatoms
   REAL(wp) ::  xpmid     !: grazing preference of microzoo for diatoms
   REAL(wp) ::  xpmepn    !: grazing preference of mesozoo for non-diatoms
   REAL(wp) ::  xpmepd    !: grazing preference of mesozoo for diatoms
   REAL(wp) ::  xpmezmi   !: grazing preference of mesozoo for microzoo
   REAL(wp) ::  xpmed     !: grazing preference of mesozoo for detritus

!! Metabolic losses
   REAL(wp) ::  xmetapn   !: non-diatom metabolic loss rate
   REAL(wp) ::  xmetapd   !: diatom     metabolic loss rate
   REAL(wp) ::  xmetazmi  !: microzoo   metabolic loss rate   
   REAL(wp) ::  xmetazme  !: mesozoo    metabolic loss rate

!! Mortality losses
   INTEGER  ::  jmpn      !: non-diatom mortality functional form
   REAL(wp) ::  xmpn      !: non-diatom mortality rate
   REAL(wp) ::  xkphn     !: non-diatom mortality half-sat constant
   INTEGER  ::  jmpd      !: diatom     mortality functional form
   REAL(wp) ::  xmpd      !: diatom     mortality rate
   REAL(wp) ::  xkphd     !: diatom     mortality half-sat constant
   INTEGER  ::  jmzmi     !: microzoo   mortality functional form
   REAL(wp) ::  xmzmi     !: microzoo   mortality rate
   REAL(wp) ::  xkzmi     !: microzoo   mortality half-sat constant
   INTEGER  ::  jmzme     !: mesozoo    mortality functional form
   REAL(wp) ::  xmzme     !: mesozoo    mortality rate   
   REAL(wp) ::  xkzme     !: mesozoo    mortality half-sat constant

!! Remineralisation
   INTEGER  ::  jmd       !: detritus T-dependent remineralisation switch
   REAL(wp) ::  xmd       !: detrital nitrogen remineralisation rate
   REAL(wp) ::  xmdc      !: detrital carbon remineralisation rate

!! Stochiometric ratios
   REAL(wp) ::  xthetap   !: phytoplankton C:N ratio
   REAL(wp) ::  xthetapn  !: non-diatom C:N ratio
                          !:  (=xthetap since xxi is the same for both groups)
   REAL(wp) ::  xthetapd  !: diatom C:N ratio
                          !:  (=xthetap since xxi is the same for both groups)
   REAL(wp) ::  xthetazmi !: microzoo C:N ratio
   REAL(wp) ::  xthetazme !: mesozoo C:N ratio
   REAL(wp) ::  xthetad   !: detritus C:N ratio
   REAL(wp) ::  xrfn      !: phytoplankton Fe:N ratio

!! Iron parameters
   INTEGER  ::  jdust     !: internal dust to soluble iron conversion switch
   REAL(wp) ::  xfe_mass  !: iron atomic mass
   REAL(wp) ::  xfe_sol   !: soluble iron fraction of aeolian dust 
   REAL(wp) ::  xfe_sed   !: sediment iron input 
   REAL(wp) ::  xLgT      !: total ligand concentration (umol/m3)
   REAL(wp) ::  xk_FeL    !: dissociation constant for (Fe + L)
   REAL(wp) ::  xk_sc_Fe  !: scavenging rate of "free" iron

!! Gravitational sinking
   REAL(wp) ::  vsed      !: detritus gravitational sinking rate
   REAL(wp) ::  xhr       !: coefficient for Martin et al. (1987) 
                          !:   remineralisation

!! Fast-sinking detritus parameters
   REAL(wp) ::  xfdfrac1  !: fast-sinking fraction of diatom nat. mort. losses
   REAL(wp) ::  xfdfrac2  !: fast-sinking fraction of diatom grazing losses
   REAL(wp) ::  xcaco3a   !: polar (high latitude) CaCO3 fraction
   REAL(wp) ::  xcaco3b   !: equatorial (low latitude) CaCO3 fraction
   REAL(wp) ::  xmassc    !: organic C mass:mole ratio, C106 H175 O40 N16 P1
   REAL(wp) ::  xmassca   !: calcium carbonate mass:mole ratio, CaCO3
   REAL(wp) ::  xmasssi   !: biogenic silicon mass:mole ratio, (H2SiO3)n
   REAL(wp) ::  xprotca   !: calcium carbonate protection ratio
   REAL(wp) ::  xprotsi   !: biogenic silicon protection ratio
   REAL(wp) ::  xfastc    !: organic C remineralisation length scale
   REAL(wp) ::  xfastca   !: calcium carbonate dissolution length scale
   REAL(wp) ::  xfastsi   !: biogenic silicon dissolution length scale

!! Miscellaneous
   REAL(wp) ::  xsdiss    !: diatom frustule dissolution rate

!! Additional parameter
   INTEGER  ::  jpkb      !: vertical layer for diagnostic of the vertical flux 

!! Optical parameters
   REAL(wp) ::   xkr0     !: water coefficient absorption in red      
   REAL(wp) ::   xkg0     !: water coefficient absorption in green   
   REAL(wp) ::   xkrp     !: pigment coefficient absorption in red   
   REAL(wp) ::   xkgp     !: pigment coefficient absorption in green 
   REAL(wp) ::   xlr      !: exposant for pigment absorption in red  
   REAL(wp) ::   xlg      !: exposant for pigment absorption in green 
   REAL(wp) ::   rpig     !: chla/chla+phea ratio                    

! Working variables 

  real(wp) qsr(jpi,jpj)         ! Surface irradiance (W/m^2) 
  real(wp) xpar(jpi,jpj,nk)     ! Layer averaged PAR (W/m^2) 
  real(wp) dust(jpi,jpj)        ! Aeolian iron deposition (gFe/m^2/month)
  real(wp) tn(jpi,jpj,nk)       ! Temperature (degC)   
  real(wp) gphit(jpi,jpj)       ! Latitude (degN)
  real(wp) ocal_ccd(jpi,jpj)    ! Carbonate compensation depth(m)
  real(wp) hmld(jpi,jpj)        ! Mixed layer depth for diagnosing mixed layer
                                !   primary production (m)
  integer mbathy(jpi,jpj)       ! Level below deepest in water column 

  INTEGER , DIMENSION(jpi,jpj) :: neln ! number of levels in the euphotic layer
  REAL(wp), DIMENSION(jpi,jpj) :: xze  ! euphotic layer depth

! Input data from external photosynthesis model

  real(wp) fjln_in(nk)
  real(wp) fjld_in(nk)
  real(wp) fchn_in(nk)
  real(wp) fchd_in(nk)

! Diagnostics (for MarMOT use)

  real(wp) zparr_top_diag(nk)
  real(wp) zparg_top_diag(nk)
  real(wp) xvpnT_diag(nk)
  real(wp) faln_diag(nk)
  real(wp) fjln_diag(nk)
  real(wp) fnln_diag(nk)
  real(wp) ffln_diag(nk)
  real(wp) fprn_diag(nk)
  real(wp) xvpdT_diag(nk)
  real(wp) fald_diag(nk)
  real(wp) fjld_diag(nk)
  real(wp) fnld_diag(nk)
  real(wp) ffld_diag(nk)
  real(wp) fsld2_diag(nk)
  real(wp) fprd_diag(nk)
  real(wp) freminn_diag(nk)
  real(wp) freminsi_diag(nk)
  real(wp) freminfe_diag(nk)
  real(wp) freminc_diag(nk)
  real(wp) freminca_diag(nk)
  real(wp) ftempn_diag(nk)
  real(wp) ftempsi_diag(nk)
  real(wp) ftempfe_diag(nk)
  real(wp) ftempc_diag(nk)
  real(wp) ftempca_diag(nk)
  real(wp) ffescav_diag(nk)

  real(wp) ffetopflx_diag
  real(wp) ffebotflx_diag

  end module medusa_globdata


