/***********************************************************************
SOFTWARE RELEASE: MarMOT 1.1.1 DATE: 23 January 2015 LICENCE: CeCILL 2.1

Includes Jan 2015 mod. to 2013 MarMOT 1.1 code (MarMOT 1.1.0) as noted 
***********************************************************************/
/***********************************************************************

SOFTWARE RELEASE: MarMOT 1.1  DATE: 21 November 2013

Copyright Natural Environment Research Council, UK

Author: John Hemmings, National Oceanography Centre 
e-mail: j.hemmings@noc.ac.uk (until 30 Nov 2013) 
Alternative e-mail for author: marmot@wessexenv.mail1.co.uk 

This software is governed by the CeCILL  license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.

***********************************************************************/

/* tabbound.c - John Hemmings (May 2001) */

/* v1.01 - Apr 2002 - Fixed field width problem in 'adjust' function 
   v1.02 - Jul 2003 - Added option to include origin 
                    - 'nintmin' defaults to 1 instead of 5 
                    - 'marginmax' & 'nintmax' defaults for invalid input are 
                      now the same as for missing input 
                    - 'nintmax' forced to be at least 10 if unspecified
   v1.03 - Jan 2015 - made return type of main function explicit */

#include "toolbox.h"

#define INFILE 1
#define OUTFILE 2
#define MARGIN 3
#define INTERVALSPEC 4
#define VOFFSET 5 

double adjust (double, double);

int main (int argc, char *argv[])

{ FILE *infile, *outfile;
  char outfname[FILENAME_MAX+1];
  char *fname;
  char invarlist[LINE_MAX+1];
  char outvarlist[LINE_MAX+1];
  char *invarname[NFIELD_MAX],*reqvarname[NFIELD_MAX],*outvarname[NFIELD_MAX];
  int byvar[NFIELD_MAX],reqvar[NFIELD_MAX],zeroflag[NFIELD_MAX];
  static double x1[NFIELD_MAX];
  static char fieldlist1[LINE_MAX+1];
  static char *field1[NFIELD_MAX];
  struct datarec d={x1,fieldlist1,field1,NULL};
  static double x2[NFIELD_MAX];
  static char fieldlist2[LINE_MAX+1];
  static char *field2[NFIELD_MAX];
  struct datarec bycurr={x2,fieldlist2,field2,NULL};
  static double x3[NFIELD_MAX];
  struct datarec dout={x3,NULL,NULL,NULL};
  double *min[NFIELD_MAX];
  double *max[NFIELD_MAX];
  double *interval[NFIELD_MAX];
  double *x[NFIELD_MAX];
  double range;
  float marginmin,marginmax,nintmin,nintmax;
  int ninvar,noutvar,nreqvar,nbyvar,byoffset,found,status,dummyflag,len,i,j,c,
      ttb,lineno,first=1,fin=0,numerror;
  char *s;

  version("1.03",VHDR,argc,argv);

  if (argc<5)
    { fprintf(stderr,
        "Usage: %s infile outfile margin_min_%%[,margin_max_%%]\n",argv[0]);
      fprintf(stderr,
        "       n_interval_min[,n_interval_max] [[0]var1 ...] [by keyvar1 ...]\n");
      exit(EXIT_FAILURE);
    }

  for (i=1;i<argc;i++)
    if (strpbrk(argv[i],WHITESPACE)!=NULL) exiterror(BADARG,argv[0],argv[i],0);
  
  s=strchr(argv[MARGIN],',');
  if (s) *(s++)='\0';

  marginmin=sgetnum(argv[MARGIN],&numerror);
  if (numerror || marginmin<0) marginmin=0;
  if (s) 
    { marginmax=sgetnum(s,&numerror);
      if (numerror) marginmax=-1;
    } 
  else marginmax=-1; 
  if (marginmax<marginmin) 
    { if (marginmin>50) marginmax=marginmin;
      else marginmax=50;
    }
  marginmin/=100;
  marginmax/=100;

  s=strchr(argv[INTERVALSPEC],',');
  if (s) *(s++)='\0';

  nintmin=sgetnum(argv[INTERVALSPEC],&numerror);
  if (numerror || nintmin<1) nintmin=1;
  if (s) 
    { nintmax=sgetnum(s,&numerror);
      if (numerror) nintmax=-1;
    } 
  else nintmax=-1;
  if (nintmax<nintmin) 
    { nintmax=2*nintmin;
      if (nintmax<10) nintmax=10;
    }

  j=VOFFSET;
  found=0;
  while(j<argc && !(found=(strcmp(argv[j],"by")==0))) j++;
  if (found) 
    { byoffset=j+1;
      nbyvar=argc-byoffset;
      nreqvar=byoffset-VOFFSET-1;
    }
  else
    { nbyvar=0;
      nreqvar=argc-VOFFSET;
    }
  for (i=0;i<nreqvar;i++) 
    { reqvarname[i]=argv[VOFFSET+i];
      if (*reqvarname[i]=='0') 
        { zeroflag[i]=1;
          reqvarname[i]++;
        }
      else zeroflag[i]=0;
    }
 
  if (strcmp(argv[INFILE],"-")==0) infile=stdin;
  else if ((infile=fopen(argv[INFILE],"r"))==NULL) 
    exiterror(OPENINERR,argv[0],argv[INFILE],0);

  status=readvar(infile,invarlist,invarname,&ninvar,&ttb,&lineno);
  if (status!=OK) exiterror(status,argv[0],argv[INFILE],lineno);

  if (indexlist(nreqvar,reqvarname,ninvar,invarname,reqvar)<nreqvar)
    { for (i=0;reqvar[i]>=0;i++);
      exiterror(VARERR,argv[0],reqvarname[i],0);
    }
  if (indexlist(nbyvar,&argv[byoffset],ninvar,invarname,byvar)<nbyvar)
    { for (i=0;byvar[i]>=0;i++);
      exiterror(VARERR,argv[0],argv[byoffset+i],0);
    }  
  if (nreqvar==0)
    { for (i=0;i<ninvar;i++) 
        { j=0;
          found=0;
          while (j<nbyvar && !(found=(byvar[j]==i))) j++;
          if (!found) 
            { reqvar[nreqvar]=i; 
              reqvarname[nreqvar]=invarname[i];
              nreqvar++;
            }
        }
    }

  noutvar=nbyvar+3*nreqvar;
  if (noutvar>NFIELD_MAX) exiterror(OUTVARERR,argv[0],NULL,0);

  if (strcmp(argv[OUTFILE],"-")==0) outfile=stdout;
  else
    { if (ttb) fname=ttbname(argv[OUTFILE],outfname);
      else fname=argv[OUTFILE];
      if ((outfile=fopen(fname,"w"))==NULL)
        exiterror(OPENOUTERR,argv[0],fname,0);
    }

  j=0;
  for (i=0;i<nbyvar;i++) outvarname[j++]=argv[byoffset+i];
  len=0; 
  for (i=0;i<nreqvar;i++) len+=2*(strlen(reqvarname[i])+2); 
  if (len>LINE_MAX+1) exiterror(OUTVARERR,argv[0],NULL,0);
  s=outvarlist;
  for (i=0;i<nreqvar;i++)
    { outvarname[j++]=s;
      *(s++)='l';
      *(s++)='.';
      strcpy(s,reqvarname[i]);
      s+=strlen(s)+1;
      outvarname[j++]=s;
      *(s++)='u';
      *(s++)='.';
      strcpy(s,reqvarname[i]);
      s+=strlen(s)+1;
      outvarname[j++]=s;
      *(s++)='i';
      *(s++)='.';
      strcpy(s,reqvarname[i]);
      s+=strlen(s)+1;
    }

  status=writevar(outfile,ttb,noutvar,outvarname,NULL,DEFAULT);
  if (status!=OK) exiterror(status,argv[0],argv[OUTFILE],0);

  for (i=0;i<nreqvar;i++)
    { min[i]=&dout.x[3*(i+1)-3];
      max[i]=&dout.x[3*(i+1)-2];
      interval[i]=&dout.x[3*(i+1)-1];
      x[i]=&d.x[reqvar[i]];
    }

  while (!fin &&
         (status=readdata(infile,ttb,ninvar,nbyvar,byvar,0,NULL,&d,&lineno))
         <=OK)

    { if (status==EOF) fin=1; 

      if (!first &&
          (fin || (c=datacomp(nbyvar,&d,byvar,&bycurr,NULL,NUMCOMP))!=0))

        { for (i=0;i<nreqvar;i++) if (*min[i]!=NODATA && *max[i]!=NODATA)
            { range=*max[i]-*min[i]; 
              *min[i]=adjust(*min[i]-marginmin*range,*min[i]-marginmax*range);
              *max[i]=adjust(*max[i]+marginmin*range,*max[i]+marginmax*range);
              range=*max[i]-*min[i]; 
              *interval[i]=adjust(range/nintmax,range/nintmin);
            } 
          status=writedata(outfile,ttb,&bycurr,nbyvar,NULL,NULL,NULL,DEFAULT,
                           DEFAULT,0);
          if (status!=OK) exiterror(status,argv[0],argv[OUTFILE],0);
          status=writedata(outfile,ttb,&dout,3*nreqvar,NULL,NULL,NULL,DEFAULT,
                           DEFAULT,1);
          if (status!=OK) exiterror(status,argv[0],argv[OUTFILE],0);
        }

      if (!fin)

        { if (first || c!=0)

            { for (i=0;i<nreqvar;i++)
                { if (zeroflag[i]) 
                    { *min[i]=0;
                      *max[i]=0;
                    }
                  else 
                    { *min[i]=NODATA;
                      *max[i]=NODATA;
                    }
                }
              copydata(&bycurr,&d,nbyvar,byvar);
            }

          first=0;

          for (i=0;i<nreqvar;i++) 
            { if (!ttb) *x[i]=sgetnum(d.field[reqvar[i]],&dummyflag); 
              if (*x[i]!=NODATA)
                { if (*min[i]==NODATA || *x[i]<*min[i]) *min[i]=*x[i];
                  if (*max[i]==NODATA || *x[i]>*max[i]) *max[i]=*x[i];
                }
            }
        }
    }

  if (!fin) exiterror(status,argv[0],argv[INFILE],lineno);

  fclose(infile);
  fclose(outfile);
  return EXIT_SUCCESS;
}


double adjust (double xstart, double xouter)  

/* Return number with least digits going from xstart towards and not beyond 
   xouter. Make last digit a 5 if possible. 
   If one digit then round up to make it a 1 if possible (e.g. 10 instead of 9).
   xstart and xouter are first converted to same precision as output. 
   Field width set to maximum space required for 'xouter' to allow room for 
   editing.  
*/

{ char smem[SIGDIG_DEF+8],*s;
  char scopy[SIGDIG_DEF+7];
  char *p,*q,digit;
  double x,xnew;
  int len,up,singledigit,error,fw=SIGDIG_DEF+6;

  smem[0]=' ';      /* 1 character buffer zone for editing */
  s=smem+1;

  sputnum(s,fw,xouter,fw,SIGDIG_DEF); 
  xouter=sgetnum(s,&error); 
  len=sputnum(s,fw,xstart,fw,SIGDIG_DEF); 
  xstart=sgetnum(s,&error); 

  if (xouter==xstart) return xstart;
  if ((xstart*xouter)<=0) return 0;

  /* xstart and xouter have the same sign - set 'up' if absolute value is to be
     increased */

  up=(xstart>0 && xouter>xstart || xstart<0 && xouter<xstart);

  x=xstart;

  p=strpbrk(s,"eE");
  if (!p) p=s+len;
  p--;
  while (*p=='0' || *p=='.') p--;

  for (;;)

    { if (up) 

        { /* Check number of digits
             Stop if already in simplest form */

          q=p-1;
          while (*q=='0' || *q=='.') q--;
          singledigit=(*q==' ' || *q=='-');
          if (singledigit && *p=='1') return sgetnum(smem,&error);
          else
      
            { /* Round up */ 

              strcpy(scopy,s);             /* Save string in case of failure */
              digit=*p;
              *p='0';
              q=p;
              do
                { q--;
                  if (*q=='.') q--; 
                  if (*q=='-') *(q-1)='-';
                  if (*q==' ' || *q=='-') *q='1';       
                  else if (*q=='9') *q='0';
                  else (*q)++;
                }
              while (*q=='0');
            }
        }

      else

        { /* Round down */

          digit=*p;
          *p='0';
        }

      xnew=sgetnum(smem,&error); 
      if (xnew==xouter) return xnew;   /* Outermost acceptable value so stop */

      if (xouter>xstart && xnew>xouter || xouter<xstart && xnew<xouter)

        { /* Out of range - try ending in 5 instead if appropriate */ 

          if (up && digit<'5' || !up && digit>'5')

            { if (up) strcpy(s,scopy);               /* Restore saved string */
              *p='5';
              xnew=sgetnum(smem,&error);
              if (xouter>xstart && xnew<=xouter || 
                  xouter<xstart && xnew>=xouter) return xnew;
            }

          /* Still out of range so return value from previous iteration */

          return x;
        } 

      /* Still in range - try again */

      while (*p=='0' || *p=='.') p--; 
      x=xnew;
    }
}
