/***********************************************************************
SOFTWARE RELEASE: MarMOT 1.1.1 DATE: 23 January 2015 LICENCE: CeCILL 2.1

Includes Jan 2015 mod. to 2013 MarMOT 1.1 code (MarMOT 1.1.0) as noted 
***********************************************************************/
/***********************************************************************

SOFTWARE RELEASE: MarMOT 1.1  DATE: 21 November 2013

Copyright Natural Environment Research Council, UK

Author: John Hemmings, National Oceanography Centre 
e-mail: j.hemmings@noc.ac.uk (until 30 Nov 2013) 
Alternative e-mail for author: marmot@wessexenv.mail1.co.uk 

This software is governed by the CeCILL  license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.

***********************************************************************/

/* tabsel.c - John Hemmings (Oct 2000) */

/* v1.01 - Fixed bug causing ttb data to be overwritten when RHS is a variable
         - RHS variable is now read as string from ttb file if required 
         - Interpret missing RHS as NODATA or NODATA_STR 
   v1.02 - Jan 2015 - made return type of main function explicit */

#include "toolbox.h"

#define INFILE 1 
#define OUTFILE 2 
#define VOFFSET 3

enum {CONSTANT,VARIABLE};
enum {EQ,NE,LT,GT,LE,GE};

int main (int argc, char *argv[])

{ FILE *infile, *outfile;
  char outfname[FILENAME_MAX+1];
  char *fname;
  char *infname;
  char varlist[LINE_MAX+1];
  char *varname[NFIELD_MAX];
  char *outvarname[NFIELD_MAX];
  char *testvarname[NFIELD_MAX];
  int testvar[NFIELD_MAX];
  int testvar2[NFIELD_MAX];
  int cond[NFIELD_MAX];
  int ctype[NFIELD_MAX];
  int valtype[NFIELD_MAX];
  int sreq[NFIELD_MAX];
  int strvar[NFIELD_MAX];
  static double x[NFIELD_MAX];
  static char fieldlist[LINE_MAX+1];
  static char *field[NFIELD_MAX];
  struct datarec d={x,fieldlist,field,NULL};
  static double x2[NFIELD_MAX];
  static char *field2[NFIELD_MAX];
  struct datarec constant={x2,NULL,field2,NULL};
  static double xnew[NFIELD_MAX];
  struct datarec dnew={xnew,NULL,NULL,NULL};
  int nvar,noutvar,ntest,nstrvar=0,found,status,numerr,c,i,j,all,req,ttb,lineno;
  char *s,*p;

  version("1.02",VHDR,argc,argv);

  if (argc<3)
    { fprintf(stderr,"Usage: %s [=]infile outfile [[@]var1[~[value1]] ...]\n",
              argv[0]);
      fprintf(stderr,"       where ~ is one of: = != < > <= >=\n");
      exit(EXIT_FAILURE);
    }

  for (i=1;i<argc;i++) 
    if (strpbrk(argv[i],WHITESPACE)!=NULL) exiterror(BADARG,argv[0],argv[i],0);

  if (*argv[INFILE]=='=')
    { all=1;
      infname=argv[INFILE]+1;
    }
  else
    { all=0;
      infname=argv[INFILE];
    }

  if (strcmp(infname,"-")==0) infile=stdin;
  else if ((infile=fopen(infname,"r"))==NULL)
    exiterror(OPENINERR,argv[0],infname,0);

  status=readvar(infile,varlist,varname,&nvar,&ttb,&lineno);
  if (status!=OK) exiterror(status,argv[0],infname,lineno);

  ntest=argc-VOFFSET;
  if (ntest>NFIELD_MAX) exiterror(OUTVARERR,argv[0],NULL,0);
  for (i=0;i<ntest;i++)

    { testvarname[i]=argv[i+VOFFSET];
      if (*testvarname[i]=='@') 
        { ctype[i]=STRCOMP;
          testvarname[i]++;
        }
      else ctype[i]=NUMCOMP;

      s=strstr(testvarname[i],"!=");
      if (s!=NULL) 
        { cond[i]=NE;
          *(s++)='\0';
        }
      else
        { s=strstr(testvarname[i],"<=");
          if (s!=NULL) 
            { cond[i]=LE;
              *(s++)='\0';
            }
          else
            { s=strstr(testvarname[i],">=");
              if (s!=NULL) 
                { cond[i]=GE;
                  *(s++)='\0';
                }
              else 
                { s=strchr(testvarname[i],'=');
                  if (s!=NULL) cond[i]=EQ;
                  else 
                    { s=strchr(testvarname[i],'<');
                      if (s!=NULL) cond[i]=LT;
                      else 
                        { s=strchr(testvarname[i],'>');
                          if (s!=NULL) cond[i]=GT;
                        }
                    }
                }
            }
        }

      valtype[i]=CONSTANT;
      constant.field[i]=NODATA_STR;
      constant.x[i]=NODATA;

      if (s==NULL) 
        { cond[i]=GT;
 /*         if (!ttb) ctype[i]=STRCOMP;*/
        }
      else
        { *(s++)='\0';
          if (*s=='\0')
            {/* if (!ttb) ctype[i]=STRCOMP;*/
            }
          else if (*s=='"') 
            { ctype[i]=STRCOMP; 
              p=strchr(++s,'"');
              if (p) *p='\0';
              constant.field[i]=s;
            }
          else
            { constant.x[i]=sgetnum(s,&numerr);
              if (numerr)
                { valtype[i]=VARIABLE;
                  j=0;
                  found=0;
                  while (j<nvar && !(found=(strcmp(varname[j],s)==0))) j++;
                  if (!found) exiterror(VARERR,argv[0],s,0); 
                  testvar2[i]=j;
                }
              else constant.field[i]=s;
            }
        }
    }

  found=indexlist(ntest,testvarname,nvar,varname,testvar);
  if (found<ntest)
    { for (i=0;testvar[i]>=0;i++);
      exiterror(VARERR,argv[0],testvarname[i],0);
    }

  for (i=0;i<nvar;i++) sreq[i]=0;
  for (i=0;i<ntest;i++)
    if (ctype[i]==STRCOMP) 
      { sreq[testvar[i]]=1;
        if (valtype[i]==VARIABLE) sreq[testvar2[i]]=1;
      }
  for (i=0;i<nvar;i++)
    if (sreq[i]) strvar[nstrvar++]=i; 

  noutvar=nvar;
  j=0;
  if (all)
    { if (++noutvar>NFIELD_MAX) exiterror(OUTVARERR,argv[0],NULL,0);
      outvarname[j++]="sel";
    }
  for (i=0;i<nvar;i++) outvarname[j++]=varname[i];
 
  if (strcmp(argv[OUTFILE],"-")==0) outfile=stdout;
  else
    { if (ttb) fname=ttbname(argv[OUTFILE],outfname);
      else fname=argv[OUTFILE];
      if ((outfile=fopen(fname,"w"))==NULL)
        exiterror(OPENOUTERR,argv[0],fname,0);
    }

  status=writevar(outfile,ttb,noutvar,outvarname,NULL,DEFAULT);
  if (status!=OK) exiterror(status,argv[0],argv[OUTFILE],0);

  while ((status=readdata(infile,ttb,nvar,0,NULL,nstrvar,strvar,&d,&lineno))
                ==OK)

    { req=1;
      for (i=0;i<ntest && req;i++)

        { if (valtype[i]==VARIABLE)
            c=datacomp(1,&d,&testvar[i],&d,&testvar2[i],ctype[i]);

          else if (ctype[i]==NUMCOMP && constant.x[i]==NODATA)

            { /* If constant=NODATA is specified then there is no 
                 string equivalent (unless it be the largest non-numeric
                 string). In this case datacomp is not used because 
                 it does a string comparison when both values are NODATA */
 
              if (d.x[testvar[i]]==NODATA) 
                { if (ttb) c=0;
                  else
                    { d.x[testvar[i]]=sgetnum(d.field[testvar[i]],&numerr);
                      if (numerr) c=-1;
                      else if (d.x[testvar[i]]==NODATA) c=0;
                      else c=1;
                    }
                }
              else c=1;
            }

          else 
            c=datacomp(1,&d,&testvar[i],&constant,&i,ctype[i]);

          if ((cond[i]==EQ && c!=0) ||
              (cond[i]==NE && c==0) ||
              (cond[i]==LT && c>=0) ||
              (cond[i]==GT && c<=0) ||
              (cond[i]==LE && c>0) ||
              (cond[i]==GE && c<0)) req=0;
        }

      if (all)
        { dnew.x[0]=(double)req;
          status=
            writedata(outfile,ttb,&dnew,1,NULL,NULL,NULL,DEFAULT,DEFAULT,0);
          if (status!=OK) exiterror(status,argv[0],argv[OUTFILE],0);
        }
      
      if (req || all) 
        { status=
            writedata(outfile,ttb,&d,nvar,NULL,NULL,NULL,DEFAULT,DEFAULT,1);
          if (status!=OK) exiterror(status,argv[0],argv[OUTFILE],0);
        }
    }

  if (status!=EOF) exiterror(status,argv[0],infname,lineno);

  fclose(infile);
  fclose(outfile);
  return EXIT_SUCCESS;
}  

