/***********************************************************************
SOFTWARE RELEASE: MarMOT 1.1.1 DATE: 23 January 2015 LICENCE: CeCILL 2.1

Includes Jan 2015 mod. to 2013 MarMOT 1.1 code (MarMOT 1.1.0) as noted 
***********************************************************************/
/***********************************************************************

SOFTWARE RELEASE: MarMOT 1.1  DATE: 21 November 2013

Copyright Natural Environment Research Council, UK

Author: John Hemmings, National Oceanography Centre 
e-mail: j.hemmings@noc.ac.uk (until 30 Nov 2013) 
Alternative e-mail for author: marmot@wessexenv.mail1.co.uk 

This software is governed by the CeCILL  license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.

***********************************************************************/

/* tabseq.c - John Hemmings (Oct 2000) */

#define TTBOUT 0 

/* v1.01 - Nov 2000 - Source code changed to allow switching between text and 
           binary output using TTBOUT (behaviour of tabseq is unchanged)
   v1.02 - Jan 2015 - made return type of main function explicit */

#include "toolbox.h"

#define OUTFILE 1 
#define VOFFSET 2 


int main (int argc, char *argv[])

{ FILE *outfile;
  char outfbuff[FILENAME_MAX+1];
  char *fname;
  char *varname[NFIELD_MAX];
  int count[NFIELD_MAX];
  int n[NFIELD_MAX];
  double start[NFIELD_MAX];
  double step[NFIELD_MAX];
  static double x[NFIELD_MAX];
  struct datarec d={x,NULL,NULL,NULL};
  enum {SIZE,RANGE}spectype;
  int nvar,status,numerr,fin=0,seqend,i,nchar,len,ttb;
  char *s,*next;
  double val,stopval;

  version("1.02",VHDR,argc,argv);

  if (argc<3)
    { fprintf(stderr,
"Usage: %s outfile var1,n1[,start1[,step1]] [var2... ...]\n",argv[0]);
      fprintf(stderr,
"   or: %s outfile var1=start1,stop1[,step1]] [var2... ...]\n",argv[0]);
      exit(EXIT_FAILURE);
    }

  for (i=1;i<argc;i++)
    if (strpbrk(argv[i],WHITESPACE)!=NULL) exiterror(BADARG,argv[0],argv[i],0);

  ttb=TTBOUT;
  if (strcmp(argv[OUTFILE],"-")==0) outfile=stdout;
  else if ((outfile=outfopen(argv[OUTFILE],outfbuff,&fname,&ttb))==NULL)
    exiterror(OPENOUTERR,argv[0],fname,0);

  nvar=argc-VOFFSET;
  if (nvar>NFIELD_MAX) exiterror(OUTVARERR,argv[0],NULL,0);

  for (i=0;i<nvar;i++)

    { varname[i]=argv[VOFFSET+i];
      next=strchr(varname[i],'=');
      if (next) 
        { spectype=RANGE;
          *(next++)='\0';
        }
      else 
        { next=strchr(varname[i],','); 
          if (next) 
            { spectype=SIZE;
              *(next++)='\0';
            }
          else exiterror(BADEXPR,argv[0],varname[i],0);
        }

      s=next;
      next=strchr(s,',');
      if (next) *(next++)='\0';

      if (spectype==SIZE)
        { if (sscanf(s,"%d%n",&n[i],&nchar)<1 || nchar!=strlen(s) || n[i]<0)
            exiterror(BADEXPR,argv[0],s,0);
        }
      else
        { start[i]=sgetnum(s,&numerr);
          if (numerr) exiterror(BADNUM,argv[0],s,0);
        }

      if (next)
        { s=next;
          next=strchr(s,',');
          if (next) *(next++)='\0';
          val=sgetnum(s,&numerr);
          if (numerr) exiterror(BADNUM,argv[0],s,0);
          if (spectype==SIZE) start[i]=val;
          else stopval=val;
        }
      else 
        { if (spectype==SIZE) start[i]=1;
          else exiterror(BADEXPR,argv[0],s,0);
        }

      if (next)
        { s=next;
          step[i]=sgetnum(s,&numerr);
          if (numerr) exiterror(BADNUM,argv[0],s,0);
        }
      else 
        { if (spectype==RANGE && stopval<start[i]) step[i]=-1;
          else step[i]=1;
        }

      if (spectype==RANGE) n[i]=1+floor(0.5+(stopval-start[i])/step[i]);

      count[i]=1;
      d.x[i]=start[i];
    } 

  status=writevar(outfile,ttb,nvar,varname,NULL,DEFAULT);
  if (status!=OK) exiterror(status,argv[0],argv[OUTFILE],0);

  while (!fin)

    { status=
        writedata(outfile,ttb,&d,nvar,NULL,NULL,NULL,DEFAULT,DEFAULT,1);
      if (status!=OK) exiterror(status,argv[0],argv[OUTFILE],0);

      seqend=1; 
      for (i=nvar-1; i>=0 && seqend; i--)
        { count[i]++;
          d.x[i]+=step[i];
          seqend=(count[i]>n[i]);
          if (seqend) 
            { count[i]=1;
              d.x[i]=start[i];
            }
        }
      
      fin=(i<0 && seqend); 
    }

  fclose(outfile);
  return EXIT_SUCCESS;
}
 
